==> data_pipelines_elasticsearch-1.0.0.tar.gz: data_pipelines_elasticsearch/src/Plugin/DatasetDestination/ElasticSearchDestination.php <==
<?php

declare(strict_types=1);

namespace Drupal\data_pipelines_elasticsearch\Plugin\DatasetDestination;

use Drupal\Component\Utility\NestedArray;
use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Utility\Error;
use Drupal\data_pipelines\Attribute\DatasetDestination;
use Drupal\data_pipelines\DatasetData;
use Drupal\data_pipelines\DatasetPipelinePluginManager;
use Drupal\data_pipelines\Destination\DatasetDestinationPluginBase;
use Drupal\data_pipelines\Entity\DatasetInterface;
use Drupal\data_pipelines\Entity\DestinationInterface;
use Drupal\data_pipelines_elasticsearch\Exception\CouldNotCreateIndexException;
use Elasticsearch\Client;
use Elasticsearch\ClientBuilder;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * A class for providing elasticsearch endpoint as an output.
 */
#[DatasetDestination(
  id: 'elasticsearch',
  label: new TranslatableMarkup('Elasticsearch'),
  description: new TranslatableMarkup('Writes datasets to an elasticsearch index'),
)]
class ElasticSearchDestination extends DatasetDestinationPluginBase implements ContainerFactoryPluginInterface {

  /**
   * The number of index shards.
   */
  const INDEX_SHARDS = 5;

  /**
   * The number of index replicas.
   */
  const INDEX_REPLICAS = 1;

  /**
   * The index codec.
   */
  const INDEX_CODEC = 'default';

  /**
   * The pipeline manager.
   *
   * @var \Drupal\data_pipelines\DatasetPipelinePluginManager
   */
  protected $pipelineManager;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The elasticsearch client.
   *
   * @var \Elasticsearch\Client
   */
  protected $client;

  /**
   * Constructs a new FileDestinationBase.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin_id for the plugin instance.
   * @param \Drupal\data_pipelines\DatasetPipelinePluginManager $pipeline_manager
   *   The pipeline manager.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger.
   */
  final public function __construct(array $configuration, $plugin_id, $plugin_definition, DatasetPipelinePluginManager $pipeline_manager, LoggerInterface $logger) {
    parent::__construct($configuration + $this->defaultConfiguration(), $plugin_id, $plugin_definition);
    $this->logger = $logger;
    $this->pipelineManager = $pipeline_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): static {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('plugin.manager.data_pipelines_pipeline'),
      $container->get('logger.channel.data_pipelines'),
    );
  }

  /**
   * Gets the ElasticSearch URL.
   *
   * @return string|null
   *   The URL.
   */
  public function getUrl(): ?string {
    return $this->configuration['url'];
  }

  /**
   * Gets the index prefix.
   *
   * @return string|null
   *   The index.
   */
  public function getPrefix(): ?string {
    return $this->configuration['prefix'];
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'url' => '',
      'username' => '',
      'password' => '',
      'prefix' => '',
      'chunk_size' => 100,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form['url'] = [
      '#type' => 'textfield',
      '#title' => $this->t('URL'),
      '#description' => $this->t('URL of the ElasticSearch cluster.'),
      '#default_value' => $this->configuration['url'],
      '#required' => TRUE,
    ];
    $form['username'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Username'),
      '#description' => $this->t('Username to authenticate with.'),
      '#default_value' => $this->configuration['username'],
    ];
    $form['password'] = [
      '#type' => 'password',
      '#title' => $this->t('Password'),
      '#description' => $this->t('Password to authenticate with.'),
      '#default_value' => $this->configuration['password'],
    ];
    $form['prefix'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Index Prefix'),
      '#description' => $this->t('The index prefix to use.'),
      '#default_value' => $this->configuration['prefix'],
    ];
    $form['chunk_size'] = [
      '#type' => 'number',
      '#min' => 1,
      '#max' => 1000,
      '#title' => $this->t('Chunk size'),
      '#description' => $this->t('How many records to index in each chunk'),
      '#default_value' => $this->configuration['chunk_size'],
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function viewSettings(): array {
    return [
      'data' => [
        '#markup' => new TranslatableMarkup("%url %ok<br/>Prefix: %prefix<br/>Chunk size: %chunk_size", [
          '%url' => $this->getUrl(),
          '%ok' => $this->ping() ? "✅ OK" : "❌ Failed",
          '%prefix' => $this->getPrefix(),
          '%chunk_size' => $this->configuration['chunk_size'],
        ]),
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    $values = $form_state->getValues();
    if (!UrlHelper::isValid($values['url'], TRUE)) {
      $form_state->setErrorByName('url', $this->t("URL must be a valid format."));
    }
  }

  /**
   * A method to get an Elasticsearch client.
   *
   * @return \Elasticsearch\Client
   *   The Elasticsearch client.
   *
   * @throws \Exception
   */
  public function getClient(): Client {
    if ($this->client) {
      return $this->client;
    }
    if (empty($this->getUrl())) {
      throw new \Exception("Cluster URL is required to initialise the ElasticSearch client");
    }
    $clientBuilder = ClientBuilder::create()
      ->setHosts([$this->getUrl()]);
    if (!empty($this->configuration['username'])) {
      $clientBuilder->setBasicAuthentication($this->configuration['username'], $this->configuration['password']);
    }
    $this->client = $clientBuilder->build();

    return $this->client;
  }

  /**
   * A method to ping the Elasticsearch instance for availability.
   *
   * @return bool
   *   The result of the ping.
   */
  public function ping(): bool {
    try {
      return $this->getClient()->ping();
    }
    catch (\Exception $e) {
      Error::logException($this->logger, $e);
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteDataSet(DatasetInterface $dataset, DestinationInterface $destination): bool {
    try {
      $this->getClient()->indices()->delete([
        'index' => $this->configuration['prefix'] . $dataset->getMachineName(),
      ]);
    }
    catch (\Exception $e) {
      $message = sprintf('The dataset %s could not be removed from the index due to %s', $dataset->label(), $e->getMessage());
      $this->logger->error($message);
      $this->messenger()->addError($message);
      return FALSE;
    }
    return TRUE;
  }

  /**
   * A method to create the index (if it doesn't exist).
   *
   * @param string $index_id
   *   The index ID.
   * @param array $settings
   *   The destination settings for the index.
   *
   * @throws \Drupal\data_pipelines_elasticsearch\Exception\CouldNotCreateIndexException
   *   When the index can't be created.
   */
  protected function createIndex(string $index_id, array $settings): void {
    $index_params = [
      'index' => $index_id,
      'body' => [
        'settings' => [
          'number_of_shards' => self::INDEX_SHARDS,
          'number_of_replicas' => self::INDEX_REPLICAS,
          'codec' => self::INDEX_CODEC,
        ],
      ],
    ];
    $index_params = NestedArray::mergeDeep($index_params, ['body' => $settings]);
    try {
      if ($this->getClient()->indices()->exists(['index' => $index_id])) {
        return;
      }
      $this->getClient()->indices()->create($index_params);
    }
    catch (\Exception $e) {
      Error::logException($this->logger, $e);
      throw new CouldNotCreateIndexException(sprintf("Failed to create index %s", $index_id), 0, $e);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getLastDelta(DatasetInterface $dataset): int {
    $index_id = $this->configuration['prefix'] . $dataset->getMachineName();
    $params = [];
    $params['index'] = $index_id;
    $count = 0;
    try {
      if ($this->getClient()->indices()->exists($params)) {
        $params['body'] = [
          "size" => 1,
          "sort" => [
            "@delta" => [
              "order" => "desc",
            ],
          ],
        ];
        $response = $this->getClient()->search($params);
        $count = (int) $response['hits']['hits'][0]['_source']['@delta'];
      }
    }
    catch (\Exception $e) {
      $this->logger->error("Failed to get a current count due to @message", [
        '@message' => $e->getMessage(),
      ]);
    }
    return $count;
  }

  /**
   * {@inheritdoc}
   */
  public function getProcessingChunkSize(): int {
    return (int) $this->configuration['chunk_size'];
  }

  /**
   * {@inheritdoc}
   */
  public function beginProcessing(DatasetInterface $dataset): bool {
    parent::beginProcessing($dataset);
    $index_id = $this->configuration['prefix'] . $dataset->getMachineName();
    $settings = $dataset->getPipeline()->getDestinationSettings($this->getPluginId());
    try {
      $this->createIndex($index_id, $settings);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->logger->error("The dataset data could not be indexed due to @message", [
        '@message' => $e->getMessage(),
      ]);
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function processChunk(DatasetInterface $dataset, array $chunk): bool {
    parent::processChunk($dataset, $chunk);
    $index_id = $this->configuration['prefix'] . $dataset->getMachineName();
    try {
      $bulk = [
        'index' => $index_id,
        'body' => [],
      ];
      foreach ($chunk as $delta => $item) {
        assert($item instanceof DatasetData);
        $bulk['body'][] = [
          'index' => [
            '_id' => $dataset->getMachineName() . ':' . $delta,
          ],
        ];
        $bulk['body'][] = ['@delta' => $delta] + $item->getArrayCopy();
      }
      if (count($bulk['body']) > 0) {
        $this->getClient()->bulk($bulk);
      }
      return TRUE;
    }
    catch (\Exception $e) {
      $this->logger->error("The dataset data could not be indexed due to @message", [
        '@message' => $e->getMessage(),
      ]);
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function processCleanup(DatasetInterface $dataset, array $invalidDeltas): bool {
    parent::processCleanup($dataset, $invalidDeltas);
    $index_id = $this->configuration['prefix'] . $dataset->getMachineName();
    try {
      $bulk = ['body' => []];
      foreach ($invalidDeltas as $delta) {
        $bulk['body'][] = [
          'delete' => [
            '_index' => $index_id,
            '_id' => $dataset->getMachineName() . ':' . $delta,
          ],
        ];
      }
      if (count($bulk['body']) > 0) {
        $this->getClient()->bulk($bulk);
      }
      return TRUE;
    }
    catch (\Exception $e) {
      $this->logger->error("The invalid dataset data could not be purged due to @message", [
        '@message' => $e->getMessage(),
      ]);
    }
    return FALSE;
  }

}

==> data_pipelines_elasticsearch-2.0.0.tar.gz: data_pipelines_elasticsearch/src/Plugin/DatasetDestination/ElasticSearchDestination.php <==
<?php

declare(strict_types=1);

namespace Drupal\data_pipelines_elasticsearch\Plugin\DatasetDestination;

use Drupal\Component\Utility\NestedArray;
use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Utility\Error;
use Drupal\data_pipelines\Attribute\DatasetDestination;
use Drupal\data_pipelines\DatasetData;
use Drupal\data_pipelines\DatasetPipelinePluginManager;
use Drupal\data_pipelines\Destination\DatasetDestinationPluginBase;
use Drupal\data_pipelines\Entity\DatasetInterface;
use Drupal\data_pipelines\Entity\DestinationInterface;
use Drupal\data_pipelines_elasticsearch\Exception\CouldNotCreateIndexException;
use Elastic\Elasticsearch\Client;
use Elastic\Elasticsearch\ClientBuilder;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * A class for providing elasticsearch endpoint as an output.
 */
#[DatasetDestination(
  id: 'elasticsearch',
  label: new TranslatableMarkup('Elasticsearch'),
  description: new TranslatableMarkup('Writes datasets to an elasticsearch index'),
)]
class ElasticSearchDestination extends DatasetDestinationPluginBase implements ContainerFactoryPluginInterface {

  /**
   * The number of index shards.
   */
  const INDEX_SHARDS = 5;

  /**
   * The number of index replicas.
   */
  const INDEX_REPLICAS = 1;

  /**
   * The index codec.
   */
  const INDEX_CODEC = 'default';

  /**
   * The pipeline manager.
   *
   * @var \Drupal\data_pipelines\DatasetPipelinePluginManager
   */
  protected $pipelineManager;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The elasticsearch client.
   *
   * @var \Elastic\Elasticsearch\Client
   */
  protected $client;

  /**
   * Constructs a new FileDestinationBase.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin_id for the plugin instance.
   * @param \Drupal\data_pipelines\DatasetPipelinePluginManager $pipeline_manager
   *   The pipeline manager.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger.
   */
  final public function __construct(array $configuration, $plugin_id, $plugin_definition, DatasetPipelinePluginManager $pipeline_manager, LoggerInterface $logger) {
    parent::__construct($configuration + $this->defaultConfiguration(), $plugin_id, $plugin_definition);
    $this->logger = $logger;
    $this->pipelineManager = $pipeline_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): static {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('plugin.manager.data_pipelines_pipeline'),
      $container->get('logger.channel.data_pipelines'),
    );
  }

  /**
   * Gets the ElasticSearch URL.
   *
   * @return string|null
   *   The URL.
   */
  public function getUrl(): ?string {
    return $this->configuration['url'];
  }

  /**
   * Gets the index prefix.
   *
   * @return string|null
   *   The index.
   */
  public function getPrefix(): ?string {
    return $this->configuration['prefix'];
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'url' => '',
      'username' => '',
      'password' => '',
      'prefix' => '',
      'chunk_size' => 100,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form['url'] = [
      '#type' => 'textfield',
      '#title' => $this->t('URL'),
      '#description' => $this->t('URL of the ElasticSearch cluster.'),
      '#default_value' => $this->configuration['url'],
      '#required' => TRUE,
    ];
    $form['username'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Username'),
      '#description' => $this->t('Username to authenticate with.'),
      '#default_value' => $this->configuration['username'],
    ];
    $form['password'] = [
      '#type' => 'password',
      '#title' => $this->t('Password'),
      '#description' => $this->t('Password to authenticate with.'),
      '#default_value' => $this->configuration['password'],
    ];
    $form['prefix'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Index Prefix'),
      '#description' => $this->t('The index prefix to use.'),
      '#default_value' => $this->configuration['prefix'],
    ];
    $form['chunk_size'] = [
      '#type' => 'number',
      '#min' => 1,
      '#max' => 1000,
      '#title' => $this->t('Chunk size'),
      '#description' => $this->t('How many records to index in each chunk'),
      '#default_value' => $this->configuration['chunk_size'],
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function viewSettings(): array {
    return [
      'data' => [
        '#markup' => new TranslatableMarkup("%url %ok<br/>Prefix: %prefix<br/>Chunk size: %chunk_size", [
          '%url' => $this->getUrl(),
          '%ok' => $this->ping() ? "✅ OK" : "❌ Failed",
          '%prefix' => $this->getPrefix(),
          '%chunk_size' => $this->configuration['chunk_size'],
        ]),
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    $values = $form_state->getValues();
    if (!UrlHelper::isValid($values['url'], TRUE)) {
      $form_state->setErrorByName('url', $this->t("URL must be a valid format."));
    }
  }

  /**
   * A method to get an Elasticsearch client.
   *
   * @return \Elastic\Elasticsearch\Client
   *   The Elasticsearch client.
   *
   * @throws \Elastic\Elasticsearch\Exception\AuthenticationException
   * @throws \Exception
   */
  public function getClient(): Client {
    if ($this->client) {
      return $this->client;
    }
    if (empty($this->getUrl())) {
      throw new \Exception("Cluster URL is required to initialise the ElasticSearch client");
    }
    $clientBuilder = ClientBuilder::create()
      ->setHosts([$this->getUrl()]);
    if (!empty($this->configuration['username'])) {
      $clientBuilder->setBasicAuthentication($this->configuration['username'], $this->configuration['password']);
    }
    $this->client = $clientBuilder->build();

    return $this->client;
  }

  /**
   * A method to ping the Elasticsearch instance for availability.
   *
   * @return bool
   *   The result of the ping.
   */
  public function ping(): bool {
    try {
      return $this->getClient()->ping()->asBool();
    }
    catch (\Exception $e) {
      Error::logException($this->logger, $e);
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteDataSet(DatasetInterface $dataset, DestinationInterface $destination): bool {
    try {
      $this->getClient()->indices()->delete([
        'index' => $this->configuration['prefix'] . $dataset->getMachineName(),
      ]);
    }
    catch (\Exception $e) {
      $message = sprintf('The dataset %s could not be removed from the index due to %s', $dataset->label(), $e->getMessage());
      $this->logger->error($message);
      $this->messenger()->addError($message);
      return FALSE;
    }
    return TRUE;
  }

  /**
   * A method to create the index (if it doesn't exist).
   *
   * @param string $index_id
   *   The index ID.
   * @param array $settings
   *   The destination settings for the index.
   *
   * @throws \Drupal\data_pipelines_elasticsearch\Exception\CouldNotCreateIndexException
   *   When the index can't be created.
   */
  protected function createIndex(string $index_id, array $settings): void {
    $index_params = [
      'index' => $index_id,
      'body' => [
        'settings' => [
          'number_of_shards' => self::INDEX_SHARDS,
          'number_of_replicas' => self::INDEX_REPLICAS,
          'codec' => self::INDEX_CODEC,
        ],
      ],
    ];
    $index_params = NestedArray::mergeDeep($index_params, ['body' => $settings]);
    try {
      if ($this->getClient()->indices()->exists(['index' => $index_id])->asBool()) {
        return;
      }
      $this->getClient()->indices()->create($index_params);
    }
    catch (\Exception $e) {
      Error::logException($this->logger, $e);
      throw new CouldNotCreateIndexException(sprintf("Failed to create index %s", $index_id), 0, $e);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getLastDelta(DatasetInterface $dataset): int {
    $index_id = $this->configuration['prefix'] . $dataset->getMachineName();
    $params = [];
    $params['index'] = $index_id;
    $count = 0;
    try {
      if ($this->getClient()->indices()->exists($params)->asBool()) {
        $params['body'] = [
          "size" => 1,
          "sort" => [
            "@delta" => [
              "order" => "desc",
            ],
          ],
        ];
        $response = $this->getClient()->search($params)->asArray();
        $count = (int) $response['hits']['hits'][0]['_source']['@delta'];
      }
    }
    catch (\Exception $e) {
      $this->logger->error("Failed to get a current count due to @message", [
        '@message' => $e->getMessage(),
      ]);
    }
    return $count;
  }

  /**
   * {@inheritdoc}
   */
  public function getProcessingChunkSize(): int {
    return (int) $this->configuration['chunk_size'];
  }

  /**
   * {@inheritdoc}
   */
  public function beginProcessing(DatasetInterface $dataset): bool {
    parent::beginProcessing($dataset);
    $index_id = $this->configuration['prefix'] . $dataset->getMachineName();
    $settings = $dataset->getPipeline()->getDestinationSettings($this->getPluginId());
    try {
      $this->createIndex($index_id, $settings);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->logger->error("The dataset data could not be indexed due to @message", [
        '@message' => $e->getMessage(),
      ]);
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function processChunk(DatasetInterface $dataset, array $chunk): bool {
    parent::processChunk($dataset, $chunk);
    $index_id = $this->configuration['prefix'] . $dataset->getMachineName();
    try {
      $bulk = [
        'index' => $index_id,
        'body' => [],
      ];
      foreach ($chunk as $delta => $item) {
        assert($item instanceof DatasetData);
        $bulk['body'][] = [
          'index' => [
            '_id' => $dataset->getMachineName() . ':' . $delta,
          ],
        ];
        $bulk['body'][] = ['@delta' => $delta] + $item->getArrayCopy();
      }
      if (count($bulk['body']) > 0) {
        $this->getClient()->bulk($bulk);
      }
      return TRUE;
    }
    catch (\Exception $e) {
      $this->logger->error("The dataset data could not be indexed due to @message", [
        '@message' => $e->getMessage(),
      ]);
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function processCleanup(DatasetInterface $dataset, array $invalidDeltas): bool {
    parent::processCleanup($dataset, $invalidDeltas);
    $index_id = $this->configuration['prefix'] . $dataset->getMachineName();
    try {
      $bulk = ['body' => []];
      foreach ($invalidDeltas as $delta) {
        $bulk['body'][] = [
          'delete' => [
            '_index' => $index_id,
            '_id' => $dataset->getMachineName() . ':' . $delta,
          ],
        ];
      }
      if (count($bulk['body']) > 0) {
        $this->getClient()->bulk($bulk);
      }
      return TRUE;
    }
    catch (\Exception $e) {
      $this->logger->error("The invalid dataset data could not be purged due to @message", [
        '@message' => $e->getMessage(),
      ]);
    }
    return FALSE;
  }

}
